@@ -43,7 +43,11 @@
  * \include passthrough_hp.cc
  */
 
-#define FUSE_USE_VERSION FUSE_MAKE_VERSION(3, 12)
+#define FUSE_USE_VERSION 35
+
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 #ifndef _GNU_SOURCE
 #define _GNU_SOURCE
@@ -78,9 +82,6 @@
 
 using namespace std;
 
-#define SFS_DEFAULT_THREADS "-1" // take libfuse value as default
-#define SFS_DEFAULT_CLONE_FD "0"
-
 /* We are re-using pointers to our `struct sfs_inode` and `struct
    sfs_dirp` elements as inodes and file handles. This means that we
    must be able to store pointer a pointer in both a fuse_ino_t
@@ -120,10 +121,9 @@ typedef std::unordered_map<SrcId, Inode> InodeMap;
 
 struct Inode {
     int fd {-1};
+    bool is_symlink {false};
     dev_t src_dev {0};
     ino_t src_ino {0};
-    int generation {0};
-    uint64_t nopen {0};
     uint64_t nlookup {0};
     std::mutex m;
 
@@ -148,15 +148,11 @@ struct Fs {
     Inode root;
     double timeout;
     bool debug;
-    bool debug_fuse;
-    bool foreground;
     std::string source;
     size_t blocksize;
     dev_t src_dev;
     bool nosplice;
     bool nocache;
-    size_t num_threads;
-    bool clone_fd;
 };
 static Fs fs{};
 
@@ -197,19 +193,13 @@ static void sfs_init(void *userdata, fuse_conn_info *conn) {
     if (conn->capable & FUSE_CAP_FLOCK_LOCKS)
         conn->want |= FUSE_CAP_FLOCK_LOCKS;
 
-    if (fs.nosplice) {
-        // FUSE_CAP_SPLICE_READ is enabled in libfuse3 by default,
-        // see do_init() in in fuse_lowlevel.c
-        // Just unset both, in case FUSE_CAP_SPLICE_WRITE would also get enabled
-        // by detault.
-        conn->want &= ~FUSE_CAP_SPLICE_READ;
-        conn->want &= ~FUSE_CAP_SPLICE_WRITE;
-    } else {
-        if (conn->capable & FUSE_CAP_SPLICE_WRITE)
-            conn->want |= FUSE_CAP_SPLICE_WRITE;
-        if (conn->capable & FUSE_CAP_SPLICE_READ)
-            conn->want |= FUSE_CAP_SPLICE_READ;
-    }
+    // Use splicing if supported. Since we are using writeback caching
+    // and readahead, individual requests should have a decent size so
+    // that splicing between fd's is well worth it.
+    if (conn->capable & FUSE_CAP_SPLICE_WRITE && !fs.nosplice)
+        conn->want |= FUSE_CAP_SPLICE_WRITE;
+    if (conn->capable & FUSE_CAP_SPLICE_READ && !fs.nosplice)
+        conn->want |= FUSE_CAP_SPLICE_READ;
 }
 
 
@@ -227,6 +217,28 @@ static void sfs_getattr(fuse_req_t req, fuse_ino_t ino, fuse_file_info *fi) {
 }
 
 
+#ifdef HAVE_UTIMENSAT
+static int utimensat_empty_nofollow(Inode& inode,
+                                    const struct timespec *tv) {
+    if (inode.is_symlink) {
+        /* Does not work on current kernels, but may in the future:
+           https://marc.info/?l=linux-kernel&m=154158217810354&w=2 */
+        auto res = utimensat(inode.fd, "", tv, AT_EMPTY_PATH | AT_SYMLINK_NOFOLLOW);
+        if (res == -1 && errno == EINVAL) {
+            /* Sorry, no race free way to set times on symlink. */
+            errno = EPERM;
+        }
+        return res;
+    }
+
+    char procname[64];
+    sprintf(procname, "/proc/self/fd/%i", inode.fd);
+
+    return utimensat(AT_FDCWD, procname, tv, 0);
+}
+#endif
+
+
 static void do_setattr(fuse_req_t req, fuse_ino_t ino, struct stat *attr,
                        int valid, struct fuse_file_info* fi) {
     Inode& inode = get_inode(ino);
@@ -285,9 +297,7 @@ static void do_setattr(fuse_req_t req, fuse_ino_t ino, struct stat *attr,
             res = futimens(fi->fh, tv);
         else {
 #ifdef HAVE_UTIMENSAT
-            char procname[64];
-            sprintf(procname, "/proc/self/fd/%i", ifd);
-            res = utimensat(AT_FDCWD, procname, tv, 0);
+            res = utimensat_empty_nofollow(inode, tv);
 #else
             res = -1;
             errno = EOPNOTSUPP;
@@ -351,29 +361,14 @@ static int do_lookup(fuse_ino_t parent, const char *name,
     }
     e->ino = reinterpret_cast<fuse_ino_t>(inode_p);
     Inode& inode {*inode_p};
-    e->generation = inode.generation;
 
-    if (inode.fd == -ENOENT) { // found unlinked inode
-        if (fs.debug)
-            cerr << "DEBUG: lookup(): inode " << e->attr.st_ino
-                 << " recycled; generation=" << inode.generation << endl;
-	/* fallthrough to new inode but keep existing inode.nlookup */
-    }
-
-    if (inode.fd > 0) { // found existing inode
+    if(inode.fd != -1) { // found existing inode
         fs_lock.unlock();
         if (fs.debug)
             cerr << "DEBUG: lookup(): inode " << e->attr.st_ino
-                 << " (userspace) already known; fd = " << inode.fd << endl;
+                 << " (userspace) already known." << endl;
         lock_guard<mutex> g {inode.m};
-
         inode.nlookup++;
-        if (fs.debug)
-            cerr << "DEBUG:" << __func__ << ":" << __LINE__ << " "
-                 <<  "inode " << inode.src_ino
-                 << " count " << inode.nlookup << endl;
-
-
         close(newfd);
     } else { // no existing inode
         /* This is just here to make Helgrind happy. It violates the
@@ -383,19 +378,14 @@ static int do_lookup(fuse_ino_t parent, const char *name,
         lock_guard<mutex> g {inode.m};
         inode.src_ino = e->attr.st_ino;
         inode.src_dev = e->attr.st_dev;
-
-        inode.nlookup++;
-        if (fs.debug)
-            cerr << "DEBUG:" << __func__ << ":" << __LINE__ << " "
-                 <<  "inode " << inode.src_ino
-                 << " count " << inode.nlookup << endl;
-
+        inode.is_symlink = S_ISLNK(e->attr.st_mode);
+        inode.nlookup = 1;
         inode.fd = newfd;
         fs_lock.unlock();
 
         if (fs.debug)
             cerr << "DEBUG: lookup(): created userspace inode " << e->attr.st_ino
-                 << "; fd = " << inode.fd << endl;
+                 << endl;
     }
 
     return 0;
@@ -470,6 +460,22 @@ static void sfs_symlink(fuse_req_t req, const char *link, fuse_ino_t parent,
 }
 
 
+static int linkat_empty_nofollow(Inode& inode, int dfd, const char *name) {
+    if (inode.is_symlink) {
+        auto res = linkat(inode.fd, "", dfd, name, AT_EMPTY_PATH);
+        if (res == -1 && (errno == ENOENT || errno == EINVAL)) {
+            /* Sorry, no race free way to hard-link a symlink. */
+            errno = EOPNOTSUPP;
+        }
+        return res;
+    }
+
+    char procname[64];
+    sprintf(procname, "/proc/self/fd/%i", inode.fd);
+    return linkat(AT_FDCWD, procname, dfd, name, AT_SYMLINK_FOLLOW);
+}
+
+
 static void sfs_link(fuse_req_t req, fuse_ino_t ino, fuse_ino_t parent,
                      const char *name) {
     Inode& inode = get_inode(ino);
@@ -479,9 +485,7 @@ static void sfs_link(fuse_req_t req, fuse_ino_t ino, fuse_ino_t parent,
     e.attr_timeout = fs.timeout;
     e.entry_timeout = fs.timeout;
 
-    char procname[64];
-    sprintf(procname, "/proc/self/fd/%i", inode.fd);
-    auto res = linkat(AT_FDCWD, procname, inode_p.fd, name, AT_SYMLINK_FOLLOW);
+    auto res = linkat_empty_nofollow(inode, inode_p.fd, name);
     if (res == -1) {
         fuse_reply_err(req, errno);
         return;
@@ -496,10 +500,6 @@ static void sfs_link(fuse_req_t req, fuse_ino_t ino, fuse_ino_t parent,
     {
         lock_guard<mutex> g {inode.m};
         inode.nlookup++;
-        if (fs.debug)
-            cerr << "DEBUG:" << __func__ << ":" << __LINE__ << " "
-                 <<  "inode " << inode.src_ino
-                 << " count " << inode.nlookup << endl;
     }
 
     fuse_reply_entry(req, &e);
@@ -532,33 +532,6 @@ static void sfs_rename(fuse_req_t req, fuse_ino_t parent, const char *name,
 
 static void sfs_unlink(fuse_req_t req, fuse_ino_t parent, const char *name) {
     Inode& inode_p = get_inode(parent);
-    // Release inode.fd before last unlink like nfsd EXPORT_OP_CLOSE_BEFORE_UNLINK
-    // to test reused inode numbers.
-    // Skip this when inode has an open file and when writeback cache is enabled.
-    if (!fs.timeout) {
-	    fuse_entry_param e;
-	    auto err = do_lookup(parent, name, &e);
-	    if (err) {
-		    fuse_reply_err(req, err);
-		    return;
-	    }
-	    if (e.attr.st_nlink == 1) {
-		    Inode& inode = get_inode(e.ino);
-		    lock_guard<mutex> g {inode.m};
-		    if (inode.fd > 0 && !inode.nopen) {
-			    if (fs.debug)
-				    cerr << "DEBUG: unlink: release inode " << e.attr.st_ino
-					    << "; fd=" << inode.fd << endl;
-			    lock_guard<mutex> g_fs {fs.mutex};
-			    close(inode.fd);
-			    inode.fd = -ENOENT;
-			    inode.generation++;
-		    }
-	    }
-
-        // decrease the ref which lookup above had increased
-        forget_one(e.ino, 1);
-    }
     auto res = unlinkat(inode_p.fd, name, 0);
     fuse_reply_err(req, res == -1 ? errno : 0);
 }
@@ -574,12 +547,6 @@ static void forget_one(fuse_ino_t ino, uint64_t n) {
         abort();
     }
     inode.nlookup -= n;
-
-    if (fs.debug)
-        cerr << "DEBUG:" << __func__ << ":" << __LINE__ << " "
-             <<  "inode " << inode.src_ino
-             << " count " << inode.nlookup << endl;
-
     if (!inode.nlookup) {
         if (fs.debug)
             cerr << "DEBUG: forget: cleaning up inode " << inode.src_ino << endl;
@@ -691,7 +658,7 @@ static bool is_dot_or_dotdot(const char *name) {
 
 
 static void do_readdir(fuse_req_t req, fuse_ino_t ino, size_t size,
-                    off_t offset, fuse_file_info *fi, const int plus) {
+                    off_t offset, fuse_file_info *fi, int plus) {
     auto d = get_dir_handle(fi);
     Inode& inode = get_inode(ino);
     lock_guard<mutex> g {inode.m};
@@ -736,23 +703,28 @@ static void do_readdir(fuse_req_t req, fuse_ino_t ino, size_t size,
 
         fuse_entry_param e{};
         size_t entsize;
-        if (plus) {
+        if(plus) {
             err = do_lookup(ino, entry->d_name, &e);
             if (err)
                 goto error;
             entsize = fuse_add_direntry_plus(req, p, rem, entry->d_name, &e, entry->d_off);
+
+            if (entsize > rem) {
+                if (fs.debug)
+                    cerr << "DEBUG: readdir(): buffer full, returning data. " << endl;
+                forget_one(e.ino, 1);
+                break;
+            }
         } else {
             e.attr.st_ino = entry->d_ino;
             e.attr.st_mode = entry->d_type << 12;
             entsize = fuse_add_direntry(req, p, rem, entry->d_name, &e.attr, entry->d_off);
-        }
 
-        if (entsize > rem) {
-            if (fs.debug)
-                cerr << "DEBUG: readdir(): buffer full, returning data. " << endl;
-            if (plus)
-                forget_one(e.ino, 1);
-            break;
+            if (entsize > rem) {
+                if (fs.debug)
+                    cerr << "DEBUG: readdir(): buffer full, returning data. " << endl;
+                break;
+            }
         }
 
         p += entsize;
@@ -828,13 +800,8 @@ static void sfs_create(fuse_req_t req, fuse_ino_t parent, const char *name,
         if (err == ENFILE || err == EMFILE)
             cerr << "ERROR: Reached maximum number of file descriptors." << endl;
         fuse_reply_err(req, err);
-	return;
-    }
-
-    Inode& inode = get_inode(e.ino);
-    lock_guard<mutex> g {inode.m};
-    inode.nopen++;
-    fuse_reply_create(req, &e, fi);
+    } else
+        fuse_reply_create(req, &e, fi);
 }
 
 
@@ -883,19 +850,14 @@ static void sfs_open(fuse_req_t req, fuse_ino_t ino, fuse_file_info *fi) {
         return;
     }
 
-    lock_guard<mutex> g {inode.m};
-    inode.nopen++;
     fi->keep_cache = (fs.timeout != 0);
-    fi->noflush = (fs.timeout == 0 && (fi->flags & O_ACCMODE) == O_RDONLY);
     fi->fh = fd;
     fuse_reply_open(req, fi);
 }
 
 
 static void sfs_release(fuse_req_t req, fuse_ino_t ino, fuse_file_info *fi) {
-    Inode& inode = get_inode(ino);
-    lock_guard<mutex> g {inode.m};
-    inode.nopen--;
+    (void) ino;
     close(fi->fh);
     fuse_reply_err(req, 0);
 }
@@ -1003,6 +965,12 @@ static void sfs_getxattr(fuse_req_t req, fuse_ino_t ino, const char *name,
     ssize_t ret;
     int saverr;
 
+    if (inode.is_symlink) {
+        /* Sorry, no race free way to getxattr on symlink. */
+        saverr = ENOTSUP;
+        goto out;
+    }
+
     char procname[64];
     sprintf(procname, "/proc/self/fd/%i", inode.fd);
 
@@ -1046,6 +1014,12 @@ static void sfs_listxattr(fuse_req_t req, fuse_ino_t ino, size_t size) {
     ssize_t ret;
     int saverr;
 
+    if (inode.is_symlink) {
+        /* Sorry, no race free way to listxattr on symlink. */
+        saverr = ENOTSUP;
+        goto out;
+    }
+
     char procname[64];
     sprintf(procname, "/proc/self/fd/%i", inode.fd);
 
@@ -1088,12 +1062,19 @@ static void sfs_setxattr(fuse_req_t req, fuse_ino_t ino, const char *name,
     ssize_t ret;
     int saverr;
 
+    if (inode.is_symlink) {
+        /* Sorry, no race free way to setxattr on symlink. */
+        saverr = ENOTSUP;
+        goto out;
+    }
+
     char procname[64];
     sprintf(procname, "/proc/self/fd/%i", inode.fd);
 
     ret = setxattr(procname, name, value, size, flags);
     saverr = ret == -1 ? errno : 0;
 
+out:
     fuse_reply_err(req, saverr);
 }
 
@@ -1104,10 +1085,17 @@ static void sfs_removexattr(fuse_req_t req, fuse_ino_t ino, const char *name) {
     ssize_t ret;
     int saverr;
 
+    if (inode.is_symlink) {
+        /* Sorry, no race free way to setxattr on symlink. */
+        saverr = ENOTSUP;
+        goto out;
+    }
+
     sprintf(procname, "/proc/self/fd/%i", inode.fd);
     ret = removexattr(procname, name);
     saverr = ret == -1 ? errno : 0;
 
+out:
     fuse_reply_err(req, saverr);
 }
 #endif
@@ -1174,16 +1162,10 @@ static cxxopts::ParseResult parse_options(int argc, char **argv) {
     opt_parser.add_options()
         ("debug", "Enable filesystem debug messages")
         ("debug-fuse", "Enable libfuse debug messages")
-        ("foreground", "Run in foreground")
         ("help", "Print help")
         ("nocache", "Disable all caching")
         ("nosplice", "Do not use splice(2) to transfer data")
-        ("single", "Run single-threaded")
-        ("num-threads", "Number of libfuse worker threads",
-                        cxxopts::value<int>()->default_value(SFS_DEFAULT_THREADS))
-        ("clone-fd", "use separate fuse device fd for each thread",
-                        cxxopts::value<bool>()->implicit_value(SFS_DEFAULT_CLONE_FD));
-
+        ("single", "Run single-threaded");
 
     // FIXME: Find a better way to limit the try clause to just
     // opt_parser.parse() (cf. https://github.com/jarro2783/cxxopts/issues/146)
@@ -1205,20 +1187,8 @@ static cxxopts::ParseResult parse_options(int argc, char **argv) {
     }
 
     fs.debug = options.count("debug") != 0;
-    fs.debug_fuse = options.count("debug-fuse") != 0;
-
-    fs.foreground = options.count("foreground") != 0;
-    if (fs.debug || fs.debug_fuse)
-        fs.foreground = true;
-
     fs.nosplice = options.count("nosplice") != 0;
-    fs.num_threads = options["num-threads"].as<int>();
-    fs.clone_fd = options["clone-fd"].as<bool>();
-    char* resolved_path = realpath(argv[1], NULL);
-    if (resolved_path == NULL)
-        warn("WARNING: realpath() failed with");
-    fs.source = std::string {resolved_path};
-    free(resolved_path);
+    fs.source = std::string {realpath(argv[1], NULL)};
 
     return options;
 }
@@ -1240,12 +1210,9 @@ static void maximize_fd_limit() {
 
 int main(int argc, char *argv[]) {
 
-    struct fuse_loop_config *loop_config = NULL;
-
     // Parse command line options
     auto options {parse_options(argc, argv)};
 
-
     // We need an fd for every dentry in our the filesystem that the
     // kernel knows about. This is way more than most processes need,
     // so try to get rid of any resource softlimit.
@@ -1254,6 +1221,7 @@ int main(int argc, char *argv[]) {
     // Initialize filesystem root
     fs.root.fd = -1;
     fs.root.nlookup = 9999;
+    fs.root.is_symlink = false;
     fs.timeout = options.count("nocache") ? 0 : 86400.0;
 
     struct stat stat;
@@ -1273,7 +1241,7 @@ int main(int argc, char *argv[]) {
     if (fuse_opt_add_arg(&args, argv[0]) ||
         fuse_opt_add_arg(&args, "-o") ||
         fuse_opt_add_arg(&args, "default_permissions,fsname=hpps") ||
-        (fs.debug_fuse && fuse_opt_add_arg(&args, "-odebug")))
+        (options.count("debug-fuse") && fuse_opt_add_arg(&args, "-odebug")))
         errx(3, "ERROR: Out of memory");
 
     fuse_lowlevel_ops sfs_oper {};
@@ -1288,21 +1256,16 @@ int main(int argc, char *argv[]) {
     // Don't apply umask, use modes exactly as specified
     umask(0);
 
-    fuse_daemonize(fs.foreground);
-
     // Mount and run main loop
-    loop_config = fuse_loop_cfg_create();
-
-    if (fs.num_threads != -1)
-        fuse_loop_cfg_set_idle_threads(loop_config, fs.num_threads);
-
+    struct fuse_loop_config loop_config;
+    loop_config.clone_fd = 0;
+    loop_config.max_idle_threads = 10;
     if (fuse_session_mount(se, argv[2]) != 0)
         goto err_out3;
     if (options.count("single"))
         ret = fuse_session_loop(se);
     else
-        ret = fuse_session_loop_mt(se, loop_config);
-
+        ret = fuse_session_loop_mt(se, &loop_config);
 
     fuse_session_unmount(se);
 
@@ -1311,8 +1274,6 @@ err_out3:
 err_out2:
     fuse_session_destroy(se);
 err_out1:
-
-    fuse_loop_cfg_destroy(loop_config);
     fuse_opt_free_args(&args);
 
     return ret ? 1 : 0;
